@@ -5,6 +5,11 @@ abstract class PFAHandler {
      * public variables
      */
 
+    /**
+     * @var array
+     */
+    public $result = array();
+
     /**
      * @var array of error messages - if a method returns false, you'll find the error message(s) here
      */
@@ -28,18 +33,18 @@ abstract class PFAHandler {
      * @var string (default) name of the database table
      * (can be overridden by $CONF[database_prefix] and $CONF[database_tables][*] via table_by_key())
      */
-    protected $db_table = null;
+    protected $db_table = '';
 
     /**
      * @var string field containing the ID
      */
-    protected $id_field = null;
+    protected $id_field = '';
 
     /**
      * @var string  field containing the label
      * defaults to $id_field if not set
      */
-    protected $label_field = null;
+    protected $label_field;
 
     /**
      * field(s) to use in the ORDER BY clause
@@ -47,7 +52,7 @@ abstract class PFAHandler {
      * defaults to $id_field if not set
      * @var string
      */
-    protected $order_by = null;
+    protected $order_by = '';
 
     /**
      * @var string
@@ -92,7 +97,10 @@ abstract class PFAHandler {
     # will be set to 0 if $admin_username is set and is not a superadmin
     protected $is_superadmin = 1;
 
-    # if set, switch to user (non-admin) mode
+    /**
+     * @var string $username
+     * if set, switch to user (non-admin) mode
+     */
     protected $username = '';
 
     # will be set to 0 if a user (non-admin) is logged in
@@ -154,12 +162,18 @@ abstract class PFAHandler {
 
     /**
      * Constructor: fill $struct etc.
-     * @param integer - 0 is edit mode, set to 1 to switch to create mode
-     * @param string - if an admin_username is specified, permissions will be restricted to the domains this admin may manage
-     * @param integer - 0 if logged in as user, 1 if logged in as admin or superadmin
+     * @param int $new - 0 is edit mode, set to 1 to switch to create mode
+     * @param string $username - if an admin_username is specified, permissions will be restricted to the domains this admin may manage
+     * @param int $is_admin - 0 if logged in as user, 1 if logged in as admin or superadmin
      */
     public function __construct($new = 0, $username = "", $is_admin = 1) {
         # set label_field if not explicitely set
+        if (empty($this->id_field)) {
+            throw new \InvalidArgumentException("id_field must be defined");
+        }
+        if (empty($this->db_table)) {
+            throw new \InvalidArgumentException("db_table must be defined");
+        }
         if (empty($this->label_field)) {
             $this->label_field = $this->id_field;
         }
@@ -201,16 +215,22 @@ abstract class PFAHandler {
 
         $this->initStruct();
 
+        /**
+         * @psalm-suppress InvalidArrayOffset
+         */
         if (!isset($this->struct['_can_edit'])) {
-            $this->struct['_can_edit'] = pacol(0,           0,      1,      'vnum', ''                   , ''                  , '', '',
+            $this->struct['_can_edit'] = pacol(0,           0,      1,      'vnum', ''                   , ''                  , '', array(),
                 /*not_in_db*/ 0,
                 /*dont_write_to_db*/ 1,
                 /*select*/ '1 as _can_edit'
                 );
         }
 
+        /**
+         * @psalm-suppress InvalidArrayOffset
+         */
         if (!isset($this->struct['_can_delete'])) {
-            $this->struct['_can_delete'] = pacol(0,         0,      1,      'vnum', ''                   , ''                  , '', '',
+            $this->struct['_can_delete'] = pacol(0,         0,      1,      'vnum', ''                   , ''                  , '', array(),
                 /*not_in_db*/ 0,
                 /*dont_write_to_db*/ 1,
                 /*select*/ '1 as _can_delete'
@@ -218,7 +238,7 @@ abstract class PFAHandler {
         }
 
         $struct_hook = Config::read($this->db_table . '_struct_hook');
-        if ($struct_hook != 'NO' && function_exists($struct_hook)) {
+        if (!empty($struct_hook) && is_string($struct_hook) && $struct_hook != 'NO' && function_exists($struct_hook)) {
             $this->struct = $struct_hook($this->struct);
         }
 
@@ -317,7 +337,13 @@ abstract class PFAHandler {
      * initialize with $id and check if it is valid
      * @param string $id
      */
-    public function init($id) {
+    public function init(string $id) : bool {
+
+        // postfix treats address lookups (aliases, mailboxes) as if they were lowercase.
+        // MySQL is normally case insenstive, PostgreSQL is case sensitive.
+        // http://www.postfix.org/aliases.5.html
+        // http://www.postfix.org/virtual.8.html
+
         $this->id = strtolower($id);
         $this->label = $this->id;
 
@@ -392,7 +418,7 @@ abstract class PFAHandler {
      * @return bool - true if all values are valid, otherwise false
      * error messages (if any) are stored in $this->errormsg
      */
-    public function set($values) {
+    public function set(array $values) {
         if (!$this->can_edit) {
             $this->errormsg[] = Config::Lang_f('edit_not_allowed', $this->label);
             return false;
@@ -495,41 +521,49 @@ abstract class PFAHandler {
      * can be used to update additional columns etc.
      * hint: modify $this->values and $this->errormsg directly as needed
      */
-    protected function setmore($values) {
+    protected function setmore(array $values) {
         # do nothing
     }
 
     /**
-     * store $this->values in the database
+     * save $this->values to the database
      *
      * converts values based on $this->struct[*][type] (boolean, password encryption)
      *
-     * calls $this->storemore() where additional things can be done
+     * calls $this->postSave() where additional things can be done
      * @return bool - true if all values were stored in the database, otherwise false
-     * error messages (if any) are stored in $this->errormsg
+     *     error messages (if any) are stored in $this->errormsg
      */
-    public function store() {
+    public function save() : bool {
+        # backwards compability: save() was once (up to 3.2.x) named store(). If a child class still uses the old name, let it override save().
+        if (method_exists($this, 'store')) {
+            error_log('store() is deprecated, please rename it to save()');
+            return $this->store();
+        }
+
         if ($this->values_valid == false) {
             $this->errormsg[] = "one or more values are invalid!";
             return false;
         }
 
-        if (!$this->beforestore()) {
+        if (!$this->preSave()) {
             return false;
         }
 
         $db_values = $this->values;
 
-        foreach (array_keys($db_values) as $key) {
+        foreach ($db_values as $key => $val) {
             switch ($this->struct[$key]['type']) { # modify field content for some types
                 case 'bool':
-                    $db_values[$key] = db_get_boolean($db_values[$key]);
+                    $val = (string) $val;
+                    $db_values[$key] = db_get_boolean($val);
                     break;
                 case 'pass':
-                    $db_values[$key] = pacrypt($db_values[$key]);
+                    $val = (string) $val;
+                    $db_values[$key] = pacrypt($val);
                     break;
                 case 'b64p':
-                    $db_values[$key] = base64_encode($db_values[$key]);
+                    $db_values[$key] = base64_encode($val);
                     break;
                 case 'quot':
                 case 'vnum':
@@ -545,19 +579,20 @@ abstract class PFAHandler {
             } # remove 'dont_write_to_db' columns
         }
 
-        if ($this->new) {
-            $result = db_insert($this->db_table, $db_values);
-        } else {
-            $result = db_update($this->db_table, $this->id_field, $this->id, $db_values);
-        }
-        if ($result != 1) {
+        try {
+            if ($this->new) {
+                $result = db_insert($this->db_table, $db_values,  array('created', 'modified'),true);
+            } else {
+                $result = db_update($this->db_table, $this->id_field, $this->id, $db_values, array('created', 'modified'), true);
+            }
+        } catch (PDOException $e) {
             $this->errormsg[] = Config::lang_f($this->msg['store_error'], $this->label);
             return false;
         }
 
-        $result = $this->storemore();
+        $result = $this->postSave();
 
-        # db_log() even if storemore() failed
+        # db_log() even if postSave() failed
         db_log($this->domain, $this->msg['logname'], $this->id);
 
         if ($result) {
@@ -570,18 +605,32 @@ abstract class PFAHandler {
     }
 
     /**
-     * called by $this->store() before storing the values in the database
-     * @return bool - if false, store() will abort
+     * called by $this->save() before storing the values in the database
+     * @return bool - if false, save() will abort
      */
-    protected function beforestore() {
+    protected function preSave() : bool {
+        # backwards compability: preSave() was once (up to 3.2.x) named beforestore(). If a child class still uses the old name, let it override preSave().
+        # Note: if a child class also has preSave(), it will override this function and obviously also the compability code.
+        if (method_exists($this, 'beforestore')) {
+            error_log('beforestore() is deprecated, please rename it to preSave()');
+            return $this->beforestore();
+        }
+
         return true; # do nothing, successfully ;-)
     }
 
     /**
-     * called by $this->store() after storing $this->values in the database
+     * called by $this->save() after storing $this->values in the database
      * can be used to update additional tables, call scripts etc.
      */
-    protected function storemore() {
+    protected function postSave() : bool {
+        # backwards compability: postSave() was once (up to 3.2.x) named storemore(). If a child class still uses the old name, let it override postSave().
+        # Note: if a child class also has postSave(), it will override this function and obviously also the compability code.
+        if (method_exists($this, 'storemore')) {
+            error_log('storemore() is deprecated, please rename it to postSave()');
+            return $this->storemore();
+        }
+
         return true; # do nothing, successfully ;-)
     }
 
@@ -700,13 +749,13 @@ abstract class PFAHandler {
      *
      * calls $this->read_from_db_postprocess() to postprocess the result
      *
-     * @param array or string condition -see build_select_query() for details
-     * @param array searchmode - see build_select_query() for details
-     * @param integer limit - maximum number of rows to return
-     * @param integer offset - number of first row to return
+     * @param array|string $condition -see build_select_query() for details
+     * @param array $searchmode - see build_select_query() for details
+     * @param int $limit - maximum number of rows to return
+     * @param int $offset - number of first row to return
      * @return array - rows (as associative array, with the ID as key)
      */
-    protected function read_from_db($condition, $searchmode = array(), $limit=-1, $offset=-1) {
+    protected function read_from_db($condition, $searchmode = array(), $limit=-1, $offset=-1) : array {
         $queryparts = $this->build_select_query($condition, $searchmode);
 
         $query = $queryparts['select_cols'] . $queryparts['from_where_order'];
@@ -717,22 +766,23 @@ abstract class PFAHandler {
             $query .= " LIMIT $limit OFFSET $offset ";
         }
 
-        $result = db_query($query);
-
         $db_result = array();
-        if ($result['rows'] != 0) {
-            while ($row = db_assoc($result['result'])) {
-                $db_result[$row[$this->id_field]] = $row;
-            }
+
+
+        $result = db_query_all($query);
+
+        foreach ($result as $row) {
+            $db_result[$row[$this->id_field]] = $row;
         }
 
-        $db_result = $this->read_from_db_postprocess($db_result);
-        return $db_result;
+        return $this->read_from_db_postprocess($db_result);
     }
 
     /**
      * allows to postprocess the database result
      * called by read_from_db()
+     * @param array $db_result
+     * @return array
      */
     protected function read_from_db_postprocess($db_result) {
         return $db_result;
@@ -762,15 +812,15 @@ abstract class PFAHandler {
 
     /**
      * get a list of one or more items with all values
-     * @param array or string $condition - see read_from_db for details
+     * @param array|string $condition - see read_from_db for details
      *        WARNING: will be changed to array only in the future, with an option to include a raw string inside the array
-     * @param array - modes to use if $condition is an array - see read_from_db for details
-     * @param integer limit - maximum number of rows to return
-     * @param integer offset - number of first row to return
+     * @param array $searchmode - modes to use if $condition is an array - see read_from_db for details
+     * @param int $limit - maximum number of rows to return
+     * @param int $offset - number of first row to return
      * @return bool - always true, no need to check ;-) (if $result is not an array, getList die()s)
      * The data is stored in $this->result (as array of rows, each row is an associative array of column => value)
      */
-    public function getList($condition, $searchmode = array(), $limit=-1, $offset=-1) {
+    public function getList($condition, $searchmode = array(), $limit=-1, $offset=-1) : bool {
         if (is_array($condition)) {
             $real_condition = array();
             foreach ($condition as $key => $value) {
@@ -788,64 +838,10 @@ abstract class PFAHandler {
             $real_condition = $condition;
         }
 
-        $result = $this->read_from_db($real_condition, $searchmode, $limit, $offset);
-
-        if (!is_array($result)) {
-            error_log('getList: read_from_db didn\'t return an array. table: ' . $this->db_table . ' - condition: $condition - limit: $limit - offset: $offset');
-            error_log('getList: This is most probably caused by read_from_db_postprocess()');
-            die('Unexpected error while reading from database! (Please check the error log for details, and open a bugreport)');
-        }
-
-        $this->result = $result;
+        $this->result = $this->read_from_db($real_condition, $searchmode, $limit, $offset);
         return true;
     }
 
-
-    /**
-     * Attempt to log a user in.
-     * @param string $username
-     * @param string $password
-     * @return boolean true on successful login (i.e. password matches etc)
-     */
-    public function login($username, $password) {
-        $username = escape_string($username);
-
-        $table = table_by_key($this->db_table);
-        $active = db_get_boolean(true);
-        $query = "SELECT password FROM $table WHERE " . $this->id_field . "='$username' AND active='$active'";
-
-        $result = db_query($query);
-        if ($result['rows'] == 1) {
-            $row = db_assoc($result['result']);
-            $crypt_password = pacrypt($password, $row['password']);
-
-            if ($row['password'] == $crypt_password) {
-                return true;
-            }
-        }
-        return false;
-    }
-
-    /**
-     * Generate and store a unique password reset token valid for one hour
-     * @param string $username
-     * @return false|string
-     */
-    public function getPasswordRecoveryCode($username) {
-        if ($this->init($username)) {
-            $token = generate_password();
-            $updatedRows = db_update($this->db_table, $this->id_field, $username, array(
-                'token' => pacrypt($token),
-                'token_validity' => date("Y-m-d H:i:s", strtotime('+ 1 hour')),
-            ));
-
-            if ($updatedRows == 1) {
-                return $token;
-            }
-        }
-        return false;
-    }
-
     /**
      * Verify user's one time password reset token
      * @param string $username
@@ -853,19 +849,18 @@ abstract class PFAHandler {
      * @return boolean true on success (i.e. code matches etc)
      */
     public function checkPasswordRecoveryCode($username, $token) {
-        $username = escape_string($username);
-
         $table = table_by_key($this->db_table);
         $active = db_get_boolean(true);
 
-        // Use PHP date for token check; this avoids a timezone issue between MySQL and PHP producing different dates.
-        $now = escape_string(date('Y-m-d H:i:s'));
+        $now = date('Y-m-d H:i:s');
 
-        $query = "SELECT token FROM $table WHERE " . $this->id_field . "='$username' AND token <> '' AND active='$active' AND '$now' < token_validity";
+        $query = "SELECT token FROM $table WHERE {$this->id_field} = :username AND token <> '' AND active = :active AND token_validity > :now ";
+        $values = array('username' => $username, 'active' => $active, 'now' => $now);
+
+        $result = db_query_all($query, $values);
+        if (sizeof($result) == 1) {
+            $row = $result[0];
 
-        $result = db_query($query);
-        if ($result['rows'] == 1) {
-            $row = db_assoc($result['result']);
             $crypt_token = pacrypt($token, $row['token']);
 
             if ($row['token'] == $crypt_token) {
@@ -895,7 +890,7 @@ abstract class PFAHandler {
     }
 
     /**
-     * @return return value of previously called method
+     * @return mixed return value of previously called method
      */
     public function result() {
         return $this->result;
